
///<reference path="./types.d.ts" />

import * as fs from 'fs';
const frontEndApis: frontEndApis = require('./frontEndApis').frontEndApis;

interface MacroResult {
    content: string;
    doBreak: boolean;
    disconnect: boolean;
    headers: any;
}
interface UniParam {
    params: string[];
    symbols: { [key: string]: (p: UniParam) => MacroResult };
    ctx: RequestCtx;
}
interface TplSection {
    content: string;
    params: string[];
    symbols: any;
}
interface Page {
    content: string;
    status: number;
    headers: any;
}
interface InitOptions {
    path?: string;
}

class TplSection implements TplSection {
    constructor(content: string, params: string[], symbols: any) {
        this.content = content;
        this.params = params;
        this.symbols = symbols;
    }
}
class MacroResult implements MacroResult {
    constructor(content: string, doBreak = false, disconnect = false, headers: any = {}) {
        this.content = content;
        this.doBreak = doBreak;
        this.disconnect = disconnect;
        this.headers = headers;
    }
}
class UniParam implements UniParam {
    constructor(params: string[], symbols: any = {}, ctx: RequestCtx) {
        this.params = params;
        this.symbols = symbols;
        this.ctx = ctx;
    }
}
class Page implements Page {
    constructor(content: string, status: number, headers: any = {}) {
        this.content = content;
        this.status = status;
        this.headers = headers;
    }
}

function smartSize(size: number) {
    const units = ['', 'K', 'M', 'G', 'T', 'P', 'E'];
    let result = size;
    for (let i = 0; i < units.length; i++) {
        if (result < 1024) {
            let precised = Math.round(result * 10) / 10;
            return `${precised} ${units[i]}`;
        }
        result = result / 1024;
    }
}

class Template {
    _regex = {
        meta: /^([\+\^]?)(.*?)(\|.*)?$/
    };
    _raw: string;
    sections: { [name: string]: TplSection };
    symbols = {
        'files': async (p: UniParam) => new MacroResult(this.sections.files.content),
        'list': async (p: UniParam) => new MacroResult(
            (await frontEndApis.file_list({ path: p.ctx.path, sse: '0', omit: 'c' }, p.ctx)).list.map((entry: DirEntry) => {
                let new_p = new UniParam(p.params, {
                    itemUrl: (p: UniParam) => new MacroResult(encodeURI(p.ctx.path + entry.n)),
                    itemName: (p: UniParam) => new MacroResult(entry.n),
                    // TODO: Format, with macros :)
                    itemModified: (p: UniParam) => new MacroResult(entry.m ? `${entry.m.getMonth() + 1}/${entry.m.getDay()}/${entry.m.getFullYear()}` : ''),
                    itemSize: (p: UniParam) => new MacroResult(smartSize(entry.s)),
                }, p.ctx);
                let f = (entry.s === undefined ? this.symbols._folder : this.symbols._file);
                return f(new_p).content;
            }).join('')
        ),
        '_file': (p: UniParam) => new MacroResult(
            this.sections.file.content
                .replace('%item-url%', p.symbols.itemUrl(p).content)
                .replace('%item-name%', p.symbols.itemName(p).content)
                .replace('%item-modified%', p.symbols.itemModified(p).content || '-')
                .replace('%item-size%', p.symbols.itemSize(p).content || '-')
        ),
        '_folder': (p: UniParam) => new MacroResult(
            this.sections.folder.content
                .replace('%item-url%', p.symbols.itemUrl(p).content)
                .replace('%item-name%', p.symbols.itemName(p).content)
                .replace('%item-modified%', p.symbols.itemModified(p).content || '-')
                .replace('%item-size%', p.symbols.itemSize(p).content || '-')
        )
    }
    constructor(path = 'hfs.tpl') {
        this._raw = '\n' + fs.readFileSync(path, { encoding: 'utf8', flag: 'r' });
        this.sections = {};
        let raw_parts = this._raw.split('\n[');
        this.sections[''] = new TplSection(raw_parts.shift(), [], {});
        for (let raw_part of raw_parts) {
            let [meta, ..._data_parts] = raw_part.split(']\n');
            let data = _data_parts.join(']\n');
            let [, _plus, _names, _params] = meta.match(this._regex.meta);
            let plus = (_plus == '+' ? 1 : _plus == '^' ? 2 : 0),
                names = _names.split('='),
                params = _params ? _params.split('|') : [];
            for (let name of names) {
                if (this.sections[name] === undefined || !plus) this.sections[name] = new TplSection(plus ? '' : data, params, {});
                if (plus) {
                    if (plus == 1) this.sections[name].content += data;
                    else if (plus == 2) this.sections[name].content = data + this.sections[name].content;
                }
            }
        }
    }
    async getText(content: string, param: UniParam) {
        let old_content: string = null;
        while (content != old_content) {
            for (let sym in this.symbols) {
                if (content.indexOf('%' + sym + '%') >= 0) {
                    content = content.replace('%' + sym + '%', (await this.symbols[sym](param)).content);
                }
            }
            old_content = content;
        }
        return new MacroResult(content);
    }
    async getSection(name: string, param: UniParam, doParse: boolean, force: boolean) {
        if ( ! this.sections[name]) return null;
        let content = this.sections[name].content;
        return doParse ? await this.getText(content, param) : new MacroResult(content);
    }
    async serve(ctx: RequestCtx) {
        if (ctx.path.endsWith('/')) {
            let result = await this.getSection('', new UniParam([], {}, ctx), true, false);
            let page = new Page(result.content, 200, result.headers);
            ctx.status = page.status;
            ctx.set(page.headers);
            ctx.type = 'html';
            ctx.body = page.content;
        }
    }
}

module.exports = Template;
